from arkitekt.messages.postman.assign.assign_log import AssignLogMessage
from arkitekt.messages.postman.assign.bounced_assign import BouncedAssignMessage
from arkitekt.messages.postman.log import LogLevel
from arkitekt.messages.postman.assign.assign_critical import AssignCriticalMessage
from arkitekt.messages.postman.provide.bounced_provide import BouncedProvideMessage
from reaktion.events import (
    CancelEvent,
    DoneEvent,
    ErrorEvent,
    EventType,
    ReturnEvent,
    SkipEvent,
    TransitionEvent,
    YieldEvent,
)
from arkitekt.messages.postman.assign.assign_done import AssignDoneMessage
from arkitekt.messages.postman.assign.assign_yield import AssignYieldsMessage
from arkitekt.messages.postman.assign.assign_cancelled import AssignCancelledMessage
from arkitekt.messages.postman.assign.assign_return import AssignReturnMessage
from arkitekt.schema.enums import NodeType
from arkitekt.packers.utils import expand_inputs
from arkitekt.messages.postman.assign.bounced_forwarded_assign import (
    BouncedForwardedAssignMessage,
)
from arkitekt.messages.postman.reserve.reserve_transition import ReserveState
from arkitekt.schema import Node
from reaktion.engine import Engine
from typing import Dict
from fluss.schema import Graph
from arkitekt.actors.base import Actor
from arkitekt.contracts.reservation import Reservation
from fluss import diagram
from arkitekt.threadvars import assign_message
import asyncio
import logging
import json

logger = logging.getLogger(__name__)


class ReactiveGraphActor(Actor):
    async def on_reservation_transition(
        self, diagramID: str, reservation: Reservation, state: ReserveState
    ):
        await self.provide_log(
            f"Event: {json.dumps(TransitionEvent(diagram_id=diagramID, reservation=reservation.reference, state=state, message='Caused by a Transition').dict())}"
        )

    async def on_provide(self, provision: BouncedProvideMessage):
        self.graph = await Graph.asyncs.get(template=self.template)
        self.node = await Node.asyncs.get(template=self.template)
        self.engine = Engine(self.graph.diagram)

        self.diagramNodeIDReservationMap = await self.engine.generateReservationsMap(
            provision.meta.reference,
            provision.meta.context,
            self.on_reservation_transition,
        )
        logger.info(self.diagramNodeIDReservationMap)
        # Now we can start the reservations

    async def on_unprovide(self, provision: BouncedProvideMessage):
        ending_futures = [
            res.end() for id, res in self.diagramNodeIDReservationMap.items()
        ]
        cancelled_tasks = await asyncio.gather(*ending_futures, return_exceptions=True)

        for i in cancelled_tasks:
            if i is None:
                logger.info("Wonderfuly shutdown")
            else:
                logger.error("Shit is going down the drains!")

        logger.info(self.diagramNodeIDReservationMap)

    async def assign_log(
        self,
        assign: BouncedAssignMessage,
        message: str,
        level: LogLevel = LogLevel.INFO,
    ):
        message = AssignLogMessage(
            data={"message": str(message), "level": level},
            meta={**assign.meta.dict(exclude={"type"})},
        )
        await self.transport.forward(message)

    async def assign_generator(self, args, kwargs, assign):
        assert len(self.engine.argNode.data.args) == len(
            args
        ), "Received different arguments then our Engines ArgNode requires"
        nodeIdConstantsMap = self.engine.generateConstantsMap(kwargs)
        logger.info(nodeIdConstantsMap)

        outQueue = asyncio.Queue()
        loop = asyncio.get_event_loop()

        async def log(message, level: LogLevel = LogLevel.INFO):
            await self.assign_log(assign, message, level=level)

        nodeIDAtomsMap = await self.engine.generateAtoms(
            outQueue, nodeIdConstantsMap, self.diagramNodeIDReservationMap, log=log
        )

        tasks = []
        for id, atom in nodeIDAtomsMap.items():
            tasks.append(loop.create_task(atom.run()))

        await outQueue.put(
            ReturnEvent(
                diagram_id=self.engine.argNode.id, handle="returns", returns=args
            )
        )
        await outQueue.put(
            DoneEvent(diagram_id=self.engine.argNode.id, handle="returns")
        )
        initial_nodes = self.engine.getInitialNodes()
        for diagramNode in initial_nodes:
            await nodeIDAtomsMap[diagramNode.id].assign_handle(
                "args",
                ReturnEvent(
                    diagram_id=self.engine.argNode.id, handle="returns", returns=[]
                ),
            )
            await nodeIDAtomsMap[diagramNode.id].assign_handle(
                "args", DoneEvent(diagram_id=self.engine.argNode.id, handle="returns")
            )

        try:
            while True:
                event: EventType = await outQueue.get()
                await log(f"Event: {json.dumps(event.dict())}")
                logger.info(f"Assingining {event} from Mainloop")

                handle_nodes = self.engine.connectedNodesWithHandle(
                    event.diagram_id, event.handle
                )

                for handle, diagramNode in handle_nodes:
                    logger.info(f" ...  to {handle} on {diagramNode}")

                    if isinstance(event, YieldEvent) or isinstance(event, ReturnEvent):
                        if isinstance(diagramNode, diagram.ReturnNode):
                            await log(
                                f"Event: {json.dumps(YieldEvent(diagram_id=diagramNode.id, returns=event.returns, handle='returns').dict())}"
                            )
                            yield event.returns
                        else:
                            await nodeIDAtomsMap[diagramNode.id].assign_handle(
                                handle, event
                            )

                    # TODO: Handle SKip events

                    if isinstance(event, DoneEvent):
                        if isinstance(diagramNode, diagram.ReturnNode):
                            await log(
                                f"Event: {json.dumps(DoneEvent(diagram_id=diagramNode.id, handle='returns').dict())}"
                            )
                            return
                        else:
                            await nodeIDAtomsMap[diagramNode.id].assign_handle(
                                handle, event
                            )

                    if isinstance(event, ErrorEvent):
                        if isinstance(diagramNode, diagram.ReturnNode):
                            await log(
                                f"Event: {json.dumps(ErrorEvent(diagram_id=diagramNode.id, exception=event.exception, message=event.message, throwing=event.throwing, handle='returns').dict())}"
                            )
                            raise type(event.exception, (Exception,), {})(event.message)
                        else:
                            await nodeIDAtomsMap[diagramNode.id].assign_handle(
                                handle, event
                            )

        except asyncio.CancelledError as e:
            logger.error("Received Cancellation babe")

            await log(
                f"Event: {json.dumps(CancelEvent(diagram_id=self.engine.argNode.id,handle='args').dict())}"
            )
            await log(
                f"Event: {json.dumps(CancelEvent(diagram_id=self.engine.kwargNode.id,handle='args').dict())}"
            )
            await log(
                f"Event: {json.dumps(CancelEvent(diagram_id=self.engine.returnNode.id,handle='args').dict())}"
            )

            for task in tasks:
                task.cancel()

            cancelled_tasks = await asyncio.gather(*tasks, return_exceptions=True)
            for i in cancelled_tasks:
                if isinstance(i, CancelEvent):
                    await log(f"Event: {json.dumps(i.dict())}")
                else:
                    logger.error("Shit is going down the drains!")

            raise e

        except Exception as e:

            for task in tasks:
                task.cancel()

            cancelled_tasks = await asyncio.gather(*tasks, return_exceptions=True)

            for i in cancelled_tasks:
                if isinstance(i, CancelEvent):
                    await log(f"Event: {json.dumps(i.dict())}")
                else:
                    print("Shit is going down the drains!")

            raise e

    async def assign_function(self, args, kwargs, assign):
        assert len(self.engine.argNode.data.args) == len(
            args
        ), "Received different arguments then our Engines ArgNode requires"
        nodeIdConstantsMap = self.engine.generateConstantsMap(kwargs)
        logger.info(nodeIdConstantsMap)

        outQueue = asyncio.Queue()
        loop = asyncio.get_event_loop()

        async def log(message, level: LogLevel = LogLevel.INFO):
            await self.assign_log(assign, message, level=level)

        nodeIDAtomsMap = await self.engine.generateAtoms(
            outQueue, nodeIdConstantsMap, self.diagramNodeIDReservationMap, log=log
        )

        tasks = []
        for id, atom in nodeIDAtomsMap.items():
            tasks.append(loop.create_task(atom.run()))

        await outQueue.put(
            ReturnEvent(
                diagram_id=self.engine.argNode.id, handle="returns", returns=args
            )
        )
        await outQueue.put(
            DoneEvent(diagram_id=self.engine.argNode.id, handle="returns")
        )
        initial_nodes = self.engine.getInitialNodes()
        for diagramNode in initial_nodes:
            await nodeIDAtomsMap[diagramNode.id].assign_handle(
                "args",
                ReturnEvent(
                    diagram_id=self.engine.argNode.id, handle="returns", returns=[]
                ),
            )
            await nodeIDAtomsMap[diagramNode.id].assign_handle(
                "args", DoneEvent(diagram_id=self.engine.argNode.id, handle="returns")
            )

        saved_returns = None

        try:
            while True:
                event: EventType = await outQueue.get()
                await log(f"Event: {json.dumps(event.dict())}")
                logger.info(f"Assingining {event} from Mainloop")

                handle_nodes = self.engine.connectedNodesWithHandle(
                    event.diagram_id, event.handle
                )

                for handle, diagramNode in handle_nodes:
                    logger.info(f" ...  to {handle} on {diagramNode}")

                    if isinstance(event, YieldEvent) or isinstance(event, ReturnEvent):
                        if isinstance(diagramNode, diagram.ReturnNode):
                            await log(
                                f"Event: {json.dumps(YieldEvent(diagram_id=diagramNode.id, returns=event.returns, handle='returns').dict())}"
                            )

                            if saved_returns is not None:
                                await log(
                                    "We have received another event on the output (probably because there is a Generator, we are omitting the old result"
                                )

                            saved_returns = event.returns
                        else:
                            await nodeIDAtomsMap[diagramNode.id].assign_handle(
                                handle, event
                            )

                    # TODO: Handle SKip events

                    if isinstance(event, DoneEvent):
                        if isinstance(diagramNode, diagram.ReturnNode):
                            await log(
                                f"Event: {json.dumps(DoneEvent(diagram_id=diagramNode.id, handle='returns').dict())}"
                            )
                            assert (
                                saved_returns is not None
                            ), "We received a Done event before a yield event"
                            return saved_returns
                        else:
                            await nodeIDAtomsMap[diagramNode.id].assign_handle(
                                handle, event
                            )

                    if isinstance(event, ErrorEvent):
                        if isinstance(diagramNode, diagram.ReturnNode):
                            await log(
                                f"Event: {json.dumps(ErrorEvent(diagram_id=diagramNode.id, exception=event.exception, message=event.message, throwing=event.throwing, handle='returns').dict())}"
                            )
                            raise type(event.exception, (Exception,), {})(event.message)
                        else:
                            await nodeIDAtomsMap[diagramNode.id].assign_handle(
                                handle, event
                            )

        except asyncio.CancelledError as e:
            logger.error("Received Cancellation babe")

            await log(
                f"Event: {json.dumps(CancelEvent(diagram_id=self.engine.argNode.id,handle='args').dict())}"
            )
            await log(
                f"Event: {json.dumps(CancelEvent(diagram_id=self.engine.kwargNode.id,handle='args').dict())}"
            )
            await log(
                f"Event: {json.dumps(CancelEvent(diagram_id=self.engine.returnNode.id,handle='args').dict())}"
            )

            for task in tasks:
                task.cancel()

            cancelled_tasks = await asyncio.gather(*tasks, return_exceptions=True)
            for i in cancelled_tasks:
                if isinstance(i, CancelEvent):
                    await log(f"Event: {json.dumps(i.dict())}")
                else:
                    logger.error("Shit is going down the drains!")

            raise e

        except Exception as e:

            for task in tasks:
                task.cancel()

            cancelled_tasks = await asyncio.gather(*tasks, return_exceptions=True)

            for i in cancelled_tasks:
                if isinstance(i, CancelEvent):
                    await log(f"Event: {json.dumps(i.dict())}")
                else:
                    print("Shit is going down the drains!")

            raise e

    async def on_assign(self, message: BouncedForwardedAssignMessage):
        try:
            logger.info(f"Assignment received {message}")
            args, kwargs = (
                message.data.args,
                message.data.kwargs,
            )  # We are not expanding the inputs as they are just passed on

            assign_message.set(message)

            if self.node.type == NodeType.FUNCTION:
                returns = await self.assign_function(args, kwargs, message)
                assign_message.set(None)

                await self.transport.forward(
                    AssignReturnMessage(
                        data={
                            "returns": returns
                            or []  # We are not shrinking the outputs as they are already shrinked
                        },
                        meta={
                            "reference": message.meta.reference,
                            "extensions": message.meta.extensions,
                        },
                    )
                )

            if self.node.type == NodeType.GENERATOR:
                async for returns in self.assign_generator(args, kwargs, message):
                    await self.transport.forward(
                        AssignYieldsMessage(
                            data={"returns": returns or []},
                            meta={
                                "reference": message.meta.reference,
                                "extensions": message.meta.extensions,
                            },
                        )
                    )

                assign_message.set(None)

                await self.transport.forward(
                    AssignDoneMessage(
                        data={"returns": None},
                        meta={
                            "reference": message.meta.reference,
                            "extensions": message.meta.extensions,
                        },
                    )
                )

        except asyncio.CancelledError as e:

            await self.transport.forward(
                AssignCancelledMessage(
                    data={"canceller": str(e)},
                    meta={
                        "reference": message.meta.reference,
                        "extensions": message.meta.extensions,
                    },
                )
            )

        except Exception as e:
            logger.exception(e)
            await self.transport.forward(
                AssignCriticalMessage(
                    data={"type": e.__class__.__name__, "message": str(e)},
                    meta={
                        "reference": message.meta.reference,
                        "extensions": message.meta.extensions,
                    },
                )
            )

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.diagramNodeReservation: Dict[
            diagram.ArkitektNode.id, Reservation
        ] = {}  # A map of



from arkitekt.messages.postman.reserve.reserve_transition import ReserveState
from arkitekt.messages.postman.provide.provide_transition import ProvideState
from typing import List, Union
from pydantic.main import BaseModel


class ProvideEvent(BaseModel):
    diagram_id: str


class TransitionEvent(ProvideEvent):
    type: str = "TRANSITION"
    state: ReserveState
    message: str
    reservation: str


class Event(BaseModel):
    diagram_id: str
    handle: str
    pass


class DoneEvent(Event):
    type: str = "DONE"
    """When a Node sends a Done event it will no longer
    send any return

    Args:
        BaseModel ([type]): [description]
    """


class CancelEvent(Event):
    type: str = "CANCEL"
    """When a Node sends a Done event it will no longer
    send any return

    Args:
        BaseModel ([type]): [description]
    """


class YieldEvent(Event):
    type: str = "YIELD"
    """When a Node sends a Done event it will no longer
    send any return

    Args:
        BaseModel ([type]): [description]
    """
    returns: List


class ReturnEvent(Event):
    type: str = "RETURN"
    """When a Node sends a Done event it will no longer
    send any return

    Args:
        BaseModel ([type]): [description]
    """
    returns: List


class ErrorEvent(Event):
    type: str = "ERROR"
    throwing: str  # the node that threw the event
    exception: str
    message: str


class SkipEvent(Event):
    type: str = "SKIP"
    skipper: str  # the node that skipped a beat
    """When a Node sends a Done event it will no longer
    send any return

    Args:
        BaseModel ([type]): [description]
    """


EventType = Union[DoneEvent, YieldEvent, ReturnEvent, ErrorEvent]



from ..base import ReactiveAtom
from fluss.diagram import DiagramNode
import asyncio
from typing import Dict


class CombinatorAtom(ReactiveAtom):

    def __init__(self, diagramNode: DiagramNode, outQueue: asyncio.Queue, constants: Dict = {}, maxsize:int = 0, **kwargs) -> None:
        super().__init__(diagramNode, outQueue, constants=constants, **kwargs)
        self.arg1_queue = asyncio.Queue(maxsize=maxsize)
        self.arg2_queue = asyncio.Queue(maxsize=maxsize)


from arkitekt.messages.postman.reserve.reserve_transition import ReserveState
from reaktion.atoms.registry import get_atom_registry
from reaktion.atoms import (
    FuncNoParallelArkitektAtom,
    GenNoParallelArkitektAtom,
    WithLatestFromAtom,
)
from arkitekt.schema.enums import NodeType
from typing import Callable, Coroutine, Dict
from herre.wards.base import WardException
from fluss import diagram
from arkitekt.schema import Node
from arkitekt.messages.postman.assign.bounced_assign import Context
import asyncio
import logging

logger = logging.getLogger(__name__)


class EngineError(Exception):
    pass


class NoNodeFoundForIdentifier(EngineError):
    pass


class Engine:
    def __init__(self, di: diagram.Diagram) -> None:
        self.nodes = [node for node in di.elements if isinstance(node, diagram.Node)]
        self.edges = [edge for edge in di.elements if isinstance(edge, diagram.Edge)]

        self.nodeIDNodeMap = {node.id: node for node in self.nodes}
        self.edgeIDEdgeMap = {edge.id: edge for edge in self.edges}

        self.argNode = next(
            (node for node in self.nodes if isinstance(node, diagram.ArgNode)), None
        )  # We only want the first one
        self.kwargNode = next(
            (node for node in self.nodes if isinstance(node, diagram.KwargNode)), None
        )  # We only want the first one
        self.returnNode = next(
            (node for node in self.nodes if isinstance(node, diagram.ReturnNode)), None
        )  # We only want the first one

        self.noneIONodes = [
            node
            for node in di.elements
            if isinstance(node, diagram.Node) and not isinstance(node, diagram.IONode)
        ]
        self.arkitektNodes = [
            node for node in self.nodes if isinstance(node, diagram.ArkitektNode)
        ]
        self.reactiveNodes = [
            node for node in self.nodes if isinstance(node, diagram.ReactiveNode)
        ]

    def connectedEdges(self, node_id: str, sourceHandle: str = None):
        return [
            edge
            for edge in self.edges
            if edge.source == node_id
            and (sourceHandle is None or edge.sourceHandle == sourceHandle)
        ]

    def connectedNodes(
        self, node_id: str, sourceHandle: str = None, targetHandle: str = None
    ):
        edg_ids = [
            edge.target
            for edge in self.connectedEdges(node_id, sourceHandle)
            if targetHandle is None or edge.targetHandle == targetHandle
        ]
        nodes = [node for node in self.nodes if node.id in edg_ids]
        return nodes

    def connectedNodesWithHandle(self, node_id: str, sourceHandle: str = None):
        edges = [edge for edge in self.connectedEdges(node_id, sourceHandle)]
        return [(edge.targetHandle, self.nodeIDNodeMap[edge.target]) for edge in edges]

    def getInitialNodes(self):
        non_kwargs_edge_targets = [
            edge.target for edge in self.edges if edge.source != self.kwargNode.id
        ]
        non_generic_nodes_that_are_not_targets = [
            node
            for node in self.arkitektNodes
            if node.id not in non_kwargs_edge_targets
        ]
        return non_generic_nodes_that_are_not_targets

    def generateConstantsMap(self, kwargs):
        nodeIDConstantsMap: Dict[str, diagram.Constants] = {}

        for kwarg, value in kwargs.items():
            for kwarg_handle, node in self.connectedNodesWithHandle(
                self.kwargNode.id, f"kwarg_{kwarg}"
            ):
                kwarg = kwarg_handle[
                    len("kwarg_") :
                ]  # Trim awai kwarg_ so that key is stillt here
                nodeIDConstantsMap.setdefault(node.id, {})[kwarg] = value

        return nodeIDConstantsMap

    async def generateReservationsMap(
        self, causing_provision: str, context: Context, transition_hook: Coroutine
    ):
        diagramIDs = [node.id for node in self.arkitektNodes]
        nodeSelectors = [node.data.selector for node in self.arkitektNodes]

        nodeInstanceFutures = [
            Node.asyncs.get(id=node.data.node.id) for node in self.arkitektNodes
        ]

        exitstates = [ReserveState.ERROR, ReserveState.CANCELLED, ReserveState.CRITICAL]

        try:
            nodeInstances = await asyncio.gather(*nodeInstanceFutures)
        except WardException as e:
            raise NoNodeFoundForIdentifier(str(e)) from e

        def id_wrapper(id, function):
            async def wrapper(res, state):
                return await function(id, res, state)

            return wrapper

        reservationsContexts = [
            node.reserve(
                **selector.dict(),
                omit_on=[] if selector.crucial else [ReserveState.DISCONNECT],
                exit_on=exitstates + [ReserveState.DISCONNECT]
                if selector.crucial
                else exitstates,
                context=context,
                transition_hook=id_wrapper(id, transition_hook),
                provision=causing_provision,
            )
            for id, node, selector in zip(diagramIDs, nodeInstances, nodeSelectors)
        ]

        reservationEnterFutures = [res.start() for res in reservationsContexts]
        reservations = await asyncio.gather(*reservationEnterFutures)

        return {
            node_id: reservation
            for node_id, reservation in zip(diagramIDs, reservations)
        }

    async def generateAtoms(self, outQueue, constants={}, reservations={}, **kwargs):
        diagramIDAtomMap = {}
        registry = get_atom_registry()
        for diagramNode in self.noneIONodes:
            diagramIDAtomMap[diagramNode.id] = registry.get_atom(diagramNode._type)(
                diagramNode,
                outQueue,
                reservation=reservations.get(diagramNode.id, None),
                constant=constants.get(diagramNode.id, {}),
                **kwargs,
            )

        return diagramIDAtomMap



import asyncio
from fluss.diagram import *
from reaktion.atoms.base import Atom
from typing import Dict, Type
from reaktion.atoms.reactive.combinators.zip import ZipAtom
from herre.wards.base import BaseWard


class NoAtomRegisted(Exception):
    pass



class AtomRegistry:

    def __init__(self) -> None:
        self.typeAtomClassMap: Dict[str, Type[Atom]] = {}

    def register_atom(self, type, atomClass: Type[Atom], overwrite=False):
        assert type not in self.typeAtomClassMap or overwrite, "We cannot register another Atom for this type. If you are sure specifiy the overwrite paramter"
        self.typeAtomClassMap[type] = atomClass

    def register_defaults(self):
        from .arkitekt import GenNoParallelArkitektAtom, FuncNoParallelArkitektAtom
        from .reactive import WithLatestFromAtom, CombineLatestAtom

        self.register_atom(ArkitektGenNode._type, GenNoParallelArkitektAtom)
        self.register_atom(ArkitektFuncNode._type, FuncNoParallelArkitektAtom)

        self.register_atom(WithLatestFromNode._type, WithLatestFromAtom) 
        self.register_atom(ZipNode._type, ZipAtom) 
        self.register_atom(CombineLatestNode._type, CombineLatestNode) 

    

    def get_atom(self, type) -> Atom:
        try:
            return self.typeAtomClassMap[type]
        except KeyError as e:
            raise NoAtomRegisted(f"No Structure registered for identifier {type}") from e




ATOM_REGISTRY = None

def get_atom_registry():
    global ATOM_REGISTRY
    if not ATOM_REGISTRY:
        ATOM_REGISTRY = AtomRegistry()
        ATOM_REGISTRY.register_defaults()


    return ATOM_REGISTRY


from abc import ABC, abstractmethod

from fluss import diagram
from ..events import (
    CancelEvent,
    DoneEvent,
    ErrorEvent,
    EventType,
    ReturnEvent,
    SkipEvent,
    YieldEvent,
)
from typing import AbstractSet, Dict, List
import asyncio
import logging
from arkitekt.messages.postman.log import LogLevel
from fluss.diagram import DiagramNode

logger = logging.getLogger(__name__)


class Atom(ABC):
    def __init__(
        self,
        diagramNode: DiagramNode,
        outQueue: asyncio.Queue,
        constants: Dict = {},
        log=None,
        **kwargs,
    ) -> None:
        self.diagramNode = diagramNode
        self.constants = constants
        self.outQueue = outQueue
        self._log = log
        if self._log:
            assert asyncio.iscoroutinefunction(
                self._log
            ), "Log function must be a coroutine or none"

    async def send_cancelled_event(self, handle):
        await self.outQueue.put(
            CancelEvent(diagram_id=self.diagramNode.id, handle=handle)
        )

    async def send_yield_event(self, handle, yields: List):
        await self.outQueue.put(
            YieldEvent(diagram_id=self.diagramNode.id, handle=handle, returns=yields)
        )

    async def send_return_event(self, handle, yields: List):
        await self.outQueue.put(
            ReturnEvent(diagram_id=self.diagramNode.id, handle=handle, returns=yields)
        )

    async def send_done_event(self, handle):
        await self.outQueue.put(
            DoneEvent(diagram_id=self.diagramNode.id, handle=handle)
        )

    async def send_skip_event(self, handle):
        await self.outQueue.put(
            SkipEvent(
                diagram_id=self.diagramNode.id,
                handle=handle,
                skipper=self.diagramNode.id,
            )
        )

    async def send_error_event(self, handle, e: Exception):
        await self.outQueue.put(
            ErrorEvent(
                diagram_id=self.diagramNode.id,
                handle=handle,
                throwing=self.diagramNode.id,
                exception=str(e.__class__.__name__),
                message=str(e),
            )
        )

    async def pass_error_event(self, handle, event: ErrorEvent):
        await self.outQueue.put(
            ErrorEvent(
                diagram_id=self.diagramNode.id,
                handle=handle,
                exception=event.exception,
                throwing=event.throwing,
                message=event.message,
            )
        )

    async def pass_skip_event(self, handle, event: SkipEvent):
        await self.outQueue.put(
            SkipEvent(
                diagram_id=self.diagramNode.id, handle=handle, skipper=event.skipper
            )
        )

    async def log(self, message: str, logLevel: LogLevel = LogLevel.INFO):
        logger.info(f"{message}")

    @abstractmethod
    async def run(self):
        raise NotImplementedError("Needs to be implemented")

    @abstractmethod
    async def assign_handle(self, handle, event: EventType):
        raise NotImplementedError("Needs to be implemented")



from .base import CombinatorAtom
from fluss.diagram import DiagramNode
import asyncio
from typing import Dict
from ....events import *
from ....exceptions import NoHandleException
import logging

logger = logging.getLogger(__name__)

class WithLatestFromAtom(CombinatorAtom):


    def __init__(self, diagramNode: DiagramNode, outQueue: asyncio.Queue, constants: Dict = {}, maxsize: int = 0, **kwargs) -> None:
        super().__init__(diagramNode, outQueue, constants=constants, maxsize=maxsize, **kwargs)
        self.arg1_done = False
        self.arg2_done = False

        self.arg1_latest = [None]
        self.arg2_latest = [None]

    async def run(self):
        try:
            while True:

                if self.arg1_done:
                    await self.log("Arg 1 is done waiting for only Arg 2")
                    event = await self.arg2_queue.get()

                    if isinstance(event, DoneEvent):
                        self.arg2_done = True

                    if isinstance(event, ErrorEvent):
                       await self.pass_error_event("return1", event)
                       break

                    if isinstance(event, YieldEvent) or isinstance(event, ReturnEvent):
                        self.arg2_latest = event.returns
                        if self.arg1_latest != [None]:
                            await self.send_yield_event("return1", self.arg1_latest + self.arg2_latest)

                elif self.arg2_done:
                    await self.log("Arg 2 is done waiting for only Arg 1")
                    event = await self.arg1_queue.get()

                    if isinstance(event, DoneEvent):
                        self.arg1_done = True

                    if isinstance(event, ErrorEvent):
                        await self.pass_error_event("return1", event)
                        break

                    if isinstance(event, YieldEvent) or isinstance(event, ReturnEvent):
                        self.arg1_latest = event.returns
                        if self.arg2_latest != [None]:
                            await self.send_yield_event("return1", self.arg1_latest + self.arg2_latest)

                else:
                    await self.log("Waiting for both Tasks")
                    loop = asyncio.get_event_loop()
                    arg1_task = loop.create_task(self.arg1_queue.get())
                    arg2_task = loop.create_task(self.arg2_queue.get())
                    done, pending = await asyncio.wait(
                            [arg1_task, arg2_task],
                            return_when=asyncio.FIRST_COMPLETED
                    )

                    for x in done:
                        if x == arg1_task: 

                            event = x.result()
                            if isinstance(event, DoneEvent):
                                self.arg1_done = True

                            if isinstance(event, ErrorEvent):
                                await self.pass_error_event("return1", event)

                            if isinstance(event, YieldEvent) or isinstance(event, ReturnEvent):
                                self.arg1_latest = event.returns
                                if self.arg2_latest != [None]:
                                    await self.send_yield_event("return1", self.arg1_latest + self.arg2_latest)


                        if x == arg2_task:
                            event = x.result()

                            if isinstance(event, DoneEvent):
                                self.arg2_done = True

                            if isinstance(event, ErrorEvent):
                                await self.pass_error_event("return1", event)

                            if isinstance(event, YieldEvent) or isinstance(event, ReturnEvent):
                                self.arg2_latest = event.returns
                                if self.arg1_latest != [None]:
                                    await self.send_yield_event("return1", self.arg1_latest + self.arg2_latest)

                        

                    for i in pending:
                        i.cancel()


                    try:
                        await asyncio.gather(*pending)
                    except asyncio.CancelledError:
                        pass

                if self.arg2_done and self.arg1_done:
                    # We are completely done
                    await self.send_done_event("return1")
                    break


        except asyncio.CancelledError as e:
            await self.log("Atom got cancelled")
            return CancelEvent(diagram_id=self.diagramNode.id, handle="returns")


        except Exception  as e:
            logger.exception(e)


    async def assign_handle(self, handle, event):
        if (handle == "arg1"): 
            return await self.arg1_queue.put(event)
        if (handle == "arg2"): return await self.arg2_queue.put(event)
        raise NoHandleException(f"No handle for {handle}")






from .base import CombinatorAtom
from fluss.diagram import DiagramNode
import asyncio
from typing import Dict
from ....events import *
from ....exceptions import NoHandleException
import logging




logger = logging.getLogger(__name__)

class CombineLatestAtom(CombinatorAtom):
    """The combineLatest operator is one of the combination operators that emits the last value from each of the observable streams when the observable emits a value.
Example, if there are 3 data streams passed as an argument to the combineLatest operator, it will take the latest emitted value by each of the argument streams in that particular order.
You won’t see the result until each of the streams have emitted at least one value. This means if you have passed in 2 argument observables to the operator, the returned observable will always emit an array of 2 values. Now since the resulting array needs to have the same length, the combineLatest operator emits only once each of the streams have emitted at least once.

    Args:
        CombinatorAtom ([type]): [description]
    """


    def __init__(self, diagramNode: DiagramNode, outQueue: asyncio.Queue, constants: Dict = {}, maxsize: int = 0, **kwargs) -> None:
        super().__init__(diagramNode, outQueue, constants=constants, maxsize=maxsize, **kwargs)
        self.arg1_done = False
        self.arg2_done = False

        self.arg1_latest = [None]
        self.arg2_latest = [None]

    async def run(self):
        try:
            while True:

                if self.arg1_done:
                    await self.log("Arg 1 is done waiting for only Arg 2")
                    event = await self.arg2_queue.get()

                    if isinstance(event, DoneEvent):
                        self.arg2_done = True

                    if isinstance(event, ErrorEvent):
                       await self.pass_error_event("return1", event)
                       break

                    if isinstance(event, YieldEvent) or isinstance(event, ReturnEvent):
                        self.arg2_latest = event.returns
                        await self.send_yield_event("return1", self.arg1_latest + self.arg2_latest)

                elif self.arg2_done:
                    await self.log("Arg 2 is done waiting for only Arg 1")
                    event = await self.arg1_queue.get()

                    if isinstance(event, DoneEvent):
                        self.arg1_done = True

                    if isinstance(event, ErrorEvent):
                        await self.pass_error_event("return1", event)
                        break

                    if isinstance(event, YieldEvent) or isinstance(event, ReturnEvent):
                        self.arg1_latest = event.returns
                        await self.send_yield_event("return1", self.arg1_latest + self.arg2_latest)

                else:
                    await self.log("Waiting for both Tasks")
                    loop = asyncio.get_event_loop()
                    arg1_task = loop.create_task(self.arg1_queue.get())
                    arg2_task = loop.create_task(self.arg2_queue.get())
                    done, pending = await asyncio.wait(
                            [arg1_task, arg2_task],
                            return_when=asyncio.FIRST_COMPLETED
                    )

                    for x in done:
                        if x == arg1_task: 

                            event = x.result()
                            if isinstance(event, DoneEvent):
                                self.arg1_done = True

                            if isinstance(event, ErrorEvent):
                                await self.pass_error_event("return1", event)

                            if isinstance(event, YieldEvent) or isinstance(event, ReturnEvent):
                                self.arg1_latest = event.returns
                                await self.send_yield_event("return1", self.arg1_latest + self.arg2_latest)


                        if x == arg2_task:
                            event = x.result()

                            if isinstance(event, DoneEvent):
                                self.arg2_done = True

                            if isinstance(event, ErrorEvent):
                                await self.pass_error_event("return1", event)

                            if isinstance(event, YieldEvent) or isinstance(event, ReturnEvent):
                                self.arg2_latest = event.returns
                                await self.send_yield_event("return1", self.arg1_latest + self.arg2_latest)

                        

                    for i in pending:
                        i.cancel()


                    try:
                        await asyncio.gather(*pending)
                    except asyncio.CancelledError:
                        pass

                if self.arg2_done and self.arg1_done:
                    # We are completely done
                    await self.send_done_event("return1")
                    break


        except asyncio.CancelledError as e:
            await self.log("Atom got cancelled")
            return CancelEvent(diagram_id=self.diagramNode.id, handle="returns")


        except Exception  as e:
            logger.exception(e)



    async def assign_handle(self, handle, event):
        if (handle == "arg1"): 
            return await self.arg1_queue.put(event)
        if (handle == "arg2"): return await self.arg2_queue.put(event)
        raise NoHandleException(f"No handle for {handle}")






from asyncio.tasks import create_task
from reaktion.actor import ReactiveGraphActor
from re import template
from fluss.schema import Graph
from arkitekt.messages.postman.provide.provide_transition import (
    ProvideState,
    ProvideTransitionMessage,
)

from arkitekt.messages.postman.provide.bounced_provide import BouncedProvideMessage

from arkitekt.agents.base import AgentException
import logging
from arkitekt.agents.app import AppAgent

logger = logging.getLogger(__name__)


class ReaktionAgent(AppAgent):
    ACTOR_PENDING_MESSAGE = "Actor is Pending"

    def __init__(self, *args, strict=False, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.strict = strict

    async def on_bounced_provide(self, message: BouncedProvideMessage):
        try:
            if message.meta.reference in self.runningActors:
                if self.strict:
                    raise AgentException(
                        "Already Running Provision Received Again. Right now causing Error. Might be omitted"
                    )
                again_provided = ProvideTransitionMessage(
                    data={
                        "message": "Provision was running on this Instance. Probably a freaking race condition",
                        "state": ProvideState.ACTIVE,
                    },
                    meta={
                        "extensions": message.meta.extensions,
                        "reference": message.meta.reference,
                    },
                )
                await self.transport.forward(again_provided)
            else:
                actor = ReactiveGraphActor()
                logger.info("Created new Reactive Graph Actor")
                self.runningActors[message.meta.reference] = actor
                task = create_task(actor.arun(message, self))
                task.add_done_callback(self.on_task_done)
                self.runningTasks[message.meta.reference] = task

        except Exception as e:
            raise AgentException("No approved actors for this template") from e



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['reaktion',
 'reaktion.atoms',
 'reaktion.atoms.arkitekt',
 'reaktion.atoms.reactive',
 'reaktion.atoms.reactive.combinators']

package_data = \
{'': ['*']}

install_requires = \
['arkitekt>=0.1.111,<0.2.0', 'fluss>=0.1.19,<0.2.0']

setup_kwargs = {
    'name': 'reaktion',
    'version': '0.1.5',
    'description': '',
    'long_description': None,
    'author': 'jhnnsrs',
    'author_email': 'jhnnsrs@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from typing import Dict
from fluss.diagram import DiagramNode
from ...atoms.base import Atom
import asyncio
from arkitekt.contracts.reservation import Omitted, Reservation
from ...events import *
import logging

logger = logging.getLogger(__name__)


class GenNoParallelArkitektAtom(Atom):
    """Non parallizable Atom, function call will not be parallized if called in row"""

    def __init__(
        self,
        diagramNode: DiagramNode,
        outQueue: asyncio.Queue,
        reservation: Reservation = None,
        constants: Dict = {},
        maxsize: int = 0,
        **kwargs,
    ) -> None:
        super().__init__(diagramNode, outQueue, constants=constants, **kwargs)
        self.reservation: Reservation = reservation
        self.inqueue = asyncio.Queue(maxsize=maxsize)

    async def run(self):
        try:
            while True:
                event: EventType = await self.inqueue.get()

                if isinstance(event, DoneEvent):
                    await self.log("Node Upstream is done, we can now also rest")
                    await self.send_done_event("returns")
                    break

                if isinstance(event, SkipEvent):
                    await self.log("Node Upstream is done, we can now also rest")
                    await self.pass_skip_event("returns", event)
                    break

                if isinstance(event, ErrorEvent):
                    await self.log("Node Upstream is done, we can now also rest")
                    await self.pass_error_event("returns", event)
                    break

                if isinstance(event, YieldEvent) or isinstance(event, ReturnEvent):
                    await self.log(
                        f"Node Upstream yielded or returned {event.returns}. Assigning to Generator"
                    )

                    try:
                        async for result in self.reservation.stream_async(
                            *event.returns,
                            **self.constants,
                            bypass_shrink=True,
                            bypass_expand=True,
                        ):
                            await self.send_yield_event("returns", result)

                    except Omitted as e:
                        logger.warn(e)
                        await self.send_skip_event("returns")

                    except Exception as e:
                        await self.send_error_event("returns", e)
                        break

        except asyncio.CancelledError as e:
            await self.log("Atom got cancelled")
            return CancelEvent(diagram_id=self.diagramNode.id, handle="returns")

        except Exception as e:
            logger.exception(e)

    async def assign_handle(self, handle: str, event: EventType):
        await self.inqueue.put(event)
